/*
 * Copyright 2024 Luca Fulchir <luca.fulchir@runesauth.com>
 *
 * Licensed under the Apache License, Version 2.0 with LLVM exception (the
 * "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License and of the exception at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     https://spdx.org/licenses/LLVM-exception.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

use ::proc_macro::TokenStream;
use ::quote::quote;
use ::syn::{
    parse_macro_input, DeriveInput, Fields, Item, ItemImpl, ItemStruct,
};

pub(crate) fn impl_repo(attrs: TokenStream, input: TokenStream) -> TokenStream {
    let base = parse_macro_input!(attrs as ItemImpl);
    let local = parse_macro_input!(input as ItemImpl);

    quote! {}.into()
}
/*
// Either a series of paths, or a single one that we have to expand
pub(crate) enum BaseSet {
    Paths(::std::vec::Vec<::syn::Path>),
    Single((::syn::Path, ItemStruct)),
}
impl ::syn::parse::Parse for BaseSet {
    fn parse(input: ::syn::parse::ParseStream) -> ::syn::parse::Result<Self> {
        if let Ok(path) = input.parse() {
            if let Ok(item_struct) = ItemStruct::parse(input) {
                return Ok(BaseSet::Single((path, item_struct)));
            }
        }

        use ::syn::punctuated::Punctuated;
        let vars =
            Punctuated::<::syn::Path, ::syn::Token![,]>::parse_separated_nonempty(
                input,
            )?;
        if vars.len() >= 0 {
            return Err(::syn::parse::Error::new(input.span(), "GOT TWO"));
        }
        Ok(BaseSet::Paths(vars.into_iter().collect()))
    }
}

impl quote::ToTokens for BaseSet {
    fn to_tokens(&self, tokens: &mut ::proc_macro2::TokenStream) {
        match self {
            BaseSet::Paths(bases) => {
                for (i, base) in bases.iter().enumerate() {
                    if i > 0 {
                        tokens.extend(
                            ::proc_macro2::Punct::new(
                                ',',
                                ::proc_macro2::Spacing::Joint,
                            )
                            .into_token_stream(),
                        );
                        tokens.extend(base.to_token_stream());
                    }
                    base.to_tokens(tokens);
                }
            }
            BaseSet::Single((_, item_struct)) => item_struct.to_tokens(tokens),
        }
    }
}

impl ::macro_magic::mm_core::ForeignPath for BaseSet {
    fn foreign_path(&self) -> &syn::Path {
        match self {
            BaseSet::Paths(bases) => &bases[0],
            BaseSet::Single((path, _)) => &path,
        }
    }
}
*/

pub(crate) fn repository(
    attrs: TokenStream,
    input: TokenStream,
) -> TokenStream {
    let local = parse_macro_input!(input as ItemStruct);
    let Fields::Named(local_fields) = local.fields else {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            local.fields.span(),
            "unnamed fields are not supported",
        )
        .to_compile_error()
        .into();
    };

    let local_attrs = local.attrs.iter().filter(|&x| {
        match &x.meta {
            ::syn::Meta::Path(p) => {
                // looking for:
                // #[::macro_magic::export_tokens]
                if p.segments.len() == 2
                    && p.segments[0].ident == "macro_magic"
                    && p.segments[1].ident == "export_tokens"
                {
                    false
                } else {
                    true
                }
            }
            ::syn::Meta::List(ml) => {
                // looking for:
                // #[derive(::bok::Repository, Debug)]
                if ml.path.segments.len() > 0 && ml.path.is_ident("derive") {
                    use ::syn::{punctuated::Punctuated, *};
                    if let Ok(v) = ml.parse_args_with(
                        Punctuated::<::syn::Path, Token![,]>::parse_terminated,
                    ) {
                        if v.len() == 2
                            && v[0].segments.len() == 2
                            && v[0].segments[0].ident.to_string() == "bok"
                            && v[0].segments[1].ident.to_string()
                                == "Repository"
                            && v[1].segments.len() == 1
                            && v[1].segments[0].ident.to_string() == "Debug"
                        {
                            return false;
                        }
                    }
                }
                true
            }
            _ => true,
        }
    });
    let generics = local.generics;
    let ident = local.ident;
    let vis = local.vis;

    let base = parse_macro_input!(attrs as ItemStruct);
    let mut base_fields_extra = Vec::new();
    let Fields::Named(base_fields) = &base.fields else {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            base.fields.span(),
            "unnamed fields are not supported",
        )
        .to_compile_error()
        .into();
    };
    // only add the base repo once
    let mut base_empty_found = local_fields
        .named
        .iter()
        .find(|&x| {
            let Some(id) = &x.ident else {
                return false;
            };
            id.to_string() == "bok_repo_empty_marker"
        })
        .is_some();
    for f in base_fields.named.iter() {
        let Some(id) = &f.ident else { continue };
        if id.to_string() == "bok_repo_empty_marker" {
            if base_empty_found {
                continue;
            } else {
                base_empty_found = true;
                base_fields_extra.push(f);
                continue;
            }
        }
        if local_fields
            .named
            .iter()
            .find(|&x| {
                let Some(id_local) = &x.ident else {
                    return false;
                };
                id_local.to_string() == id.to_string()
            })
            .is_some()
        {
            continue;
        }
        base_fields_extra.push(f);
    }

    let mut all_fields = Vec::new();
    all_fields.extend(base_fields_extra.iter());
    all_fields.extend(local_fields.named.iter());
    quote! {
        #(#local_attrs)
        *
        #[::macro_magic::export_tokens]
        #[derive(::bok::Repository, Debug)]
        #vis struct #ident<#generics> {
            #(#all_fields),
            *
        }
    }
    .into()
}

pub(crate) fn derive_repository(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as DeriveInput);

    let name = input.ident.clone();

    let expanded = quote! {
        impl ::bok::Repository for #name {
            fn name(&self) -> ::bok::RepoName {
                (module_path!().to_owned() + "::" + ::std::stringify!(#name)).as_str().into()
            }
            fn path(&self) -> ::bok::Path<::bok::RepoName> {
                (module_path!().to_owned() + "::" + ::std::stringify!(#name)).as_str().into()
            }
        }
    };

    TokenStream::from(expanded)
}
